#include <Python.h>
#include <structmember.h>

/* ECA */
typedef struct {
    PyObject_HEAD
    size_t size;
    char *state;
    char *tmp_state;
    int rule[8];
} ECA;

static void ECA_dealloc(ECA* self) {
    free(self->state);
    free(self->tmp_state);
    self->ob_type->tp_free((PyObject*)self);
}

static int ECA_init(ECA *self, PyObject *args) {
    unsigned int rule;

    if (self->state != NULL) {
        free(self->state);
    }

    if (self->tmp_state != NULL) {
        free(self->tmp_state);
    }

    if (!PyArg_ParseTuple(args, "II", &self->size, &rule)) {
        return -1;
    }

    /* We're going to assume we won't be unable to allocate memory */
    self->state = calloc(self->size, sizeof(*self->state));
    self->tmp_state = calloc(self->size, sizeof(*self->tmp_state));

    self->rule[0] = !!(rule & 1);
    self->rule[1] = !!(rule & 2);
    self->rule[2] = !!(rule & 4);
    self->rule[3] = !!(rule & 8);
    self->rule[4] = !!(rule & 16);
    self->rule[5] = !!(rule & 32);
    self->rule[6] = !!(rule & 64);
    self->rule[7] = !!(rule & 128);

    return 0;
}

static PyObject *ECA_getState(ECA *self) {
    PyObject *result = PyTuple_New(self->size);

    for (int i = 0; i < self->size; i++) {
        PyTuple_SET_ITEM(result, i, PyInt_FromLong((long) self->state[i]));
    }

    return result;
}

#define STATES_TO_RULE(a, b, c) ((a << 2) + (b << 1) + c)
static PyObject *ECA_nextState(ECA *self) {
    self->tmp_state[0] = self->rule[STATES_TO_RULE(
            self->state[self->size - 1],
            self->state[0],
            self->state[1]
    )];

    for (int i = 1; i < self->size - 1; i++) {
        self->tmp_state[i] = self->rule[STATES_TO_RULE(
                self->state[i - 1],
                self->state[i],
                self->state[i + 1]
        )];
    }

    int i = self->size - 1;
    self->tmp_state[i] = self->rule[STATES_TO_RULE(
            self->state[i - 1],
            self->state[i],
            self->state[0]
    )];

    memcpy(self->state, self->tmp_state, self->size * sizeof(*self->state));

    Py_INCREF(Py_None);
    return Py_None;
}

static Py_ssize_t ECA_size(ECA *self) {
    return self->size;
}

static PyObject *ECA_getitem(ECA *self, PyObject *key) {
    long i = PyInt_AsLong(key);
    if (i > (self->size - 1)) {
        return NULL;
    }

    return PyInt_FromLong((long) self->state[i]);
}

static int ECA_setitem(ECA *self, PyObject *key, PyObject *value) {
    long i = PyInt_AsLong(key),
         n = PyInt_AsLong(value) > 0;

    if (i > (self->size - 1)) {
        return -1;
    }

    self->state[i] = n;

    return 0;
}

static PyMethodDef ECA_methods[] = {
    {"getState", (PyCFunction)ECA_getState, METH_NOARGS, "Return the current CA state."},
    {"nextState", (PyCFunction)ECA_nextState, METH_NOARGS, "Calculate the next CA state."},
    {NULL} /* Sentinel */
};

static PyMappingMethods ECA_mappings = {
    (lenfunc)ECA_size,
    (binaryfunc)ECA_getitem,
    (objobjargproc)ECA_setitem,
};

static PyTypeObject ECAType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "eca.ECA",                 /*tp_name*/
    sizeof(ECA),               /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)ECA_dealloc,   /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    &ECA_mappings,             /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,        /*tp_flags*/
    "Elementary Cellular Automaton",           /* tp_doc */
    0,                     /* tp_traverse */
    0,                     /* tp_clear */
    0,                     /* tp_richcompare */
    0,                     /* tp_weaklistoffset */
    0,                     /* tp_iter */
    0,                     /* tp_iternext */
    ECA_methods,             /* tp_methods */
    0,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)ECA_init,      /* tp_init */
    0,                         /* tp_alloc */
    PyType_GenericNew,         /* tp_new */
};

/* NUECA */
typedef struct {
    PyObject_HEAD
    size_t size;
    char *state;
    char *tmp_state;
    int rule1[8];
    int rule2[8];
} NUECA;

static void NUECA_dealloc(NUECA* self) {
    free(self->state);
    free(self->tmp_state);
    self->ob_type->tp_free((PyObject*)self);
}

static int NUECA_init(NUECA *self, PyObject *args) {
    unsigned int rule1, rule2;

    if (self->state != NULL) {
        free(self->state);
    }

    if (self->tmp_state != NULL) {
        free(self->tmp_state);
    }

    if (!PyArg_ParseTuple(args, "III", &self->size, &rule1, &rule2)) {
        return -1;
    }

    /* We're going to assume we won't be unable to allocate memory */
    self->state = calloc(self->size, sizeof(*self->state));
    self->tmp_state = calloc(self->size, sizeof(*self->tmp_state));

    self->rule1[0] = !!(rule1 & 1);
    self->rule1[1] = !!(rule1 & 2);
    self->rule1[2] = !!(rule1 & 4);
    self->rule1[3] = !!(rule1 & 8);
    self->rule1[4] = !!(rule1 & 16);
    self->rule1[5] = !!(rule1 & 32);
    self->rule1[6] = !!(rule1 & 64);
    self->rule1[7] = !!(rule1 & 128);

    self->rule2[0] = !!(rule2 & 1);
    self->rule2[1] = !!(rule2 & 2);
    self->rule2[2] = !!(rule2 & 4);
    self->rule2[3] = !!(rule2 & 8);
    self->rule2[4] = !!(rule2 & 16);
    self->rule2[5] = !!(rule2 & 32);
    self->rule2[6] = !!(rule2 & 64);
    self->rule2[7] = !!(rule2 & 128);

    return 0;
}

static PyObject *NUECA_getState(NUECA *self) {
    PyObject *result = PyTuple_New(self->size);

    for (int i = 0; i < self->size; i++) {
        PyTuple_SET_ITEM(result, i, PyInt_FromLong((long) self->state[i]));
    }

    return result;
}

#define STATES_TO_RULE(a, b, c) ((a << 2) + (b << 1) + c)
static PyObject *NUECA_nextState(NUECA *self) {
    self->tmp_state[0] = self->rule1[STATES_TO_RULE(
            self->state[self->size - 1],
            self->state[0],
            self->state[1]
    )];

    for (int i = 1; i < (self->size / 2); i++) {
        self->tmp_state[i] = self->rule1[STATES_TO_RULE(
                self->state[i - 1],
                self->state[i],
                self->state[i + 1]
        )];
    }

    for (int i = (self->size / 2); i < (self->size - 1); i++) {
        self->tmp_state[i] = self->rule2[STATES_TO_RULE(
                self->state[i - 1],
                self->state[i],
                self->state[i + 1]
        )];
    }

    int i = self->size - 1;
    self->tmp_state[i] = self->rule2[STATES_TO_RULE(
            self->state[i - 1],
            self->state[i],
            self->state[0]
    )];

    memcpy(self->state, self->tmp_state, self->size * sizeof(*self->state));

    Py_INCREF(Py_None);
    return Py_None;
}

static Py_ssize_t NUECA_size(NUECA *self) {
    return self->size;
}

static PyObject *NUECA_getitem(NUECA *self, PyObject *key) {
    long i = PyInt_AsLong(key);
    if (i > (self->size - 1)) {
        return NULL;
    }

    return PyInt_FromLong((long) self->state[i]);
}

static int NUECA_setitem(NUECA *self, PyObject *key, PyObject *value) {
    long i = PyInt_AsLong(key),
         n = PyInt_AsLong(value) > 0;

    if (i > (self->size - 1)) {
        return -1;
    }

    self->state[i] = n;

    return 0;
}

static PyMethodDef NUECA_methods[] = {
    {"getState", (PyCFunction)NUECA_getState, METH_NOARGS, "Return the current CA state."},
    {"nextState", (PyCFunction)NUECA_nextState, METH_NOARGS, "Calculate the next CA state."},
    {NULL} /* Sentinel */
};

static PyMappingMethods NUECA_mappings = {
    (lenfunc)NUECA_size,
    (binaryfunc)NUECA_getitem,
    (objobjargproc)NUECA_setitem,
};

static PyTypeObject NUECAType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "eca.NUECA",                 /*tp_name*/
    sizeof(NUECA),               /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)NUECA_dealloc,   /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    &NUECA_mappings,             /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,        /*tp_flags*/
    "Elementary Cellular Automaton",           /* tp_doc */
    0,                     /* tp_traverse */
    0,                     /* tp_clear */
    0,                     /* tp_richcompare */
    0,                     /* tp_weaklistoffset */
    0,                     /* tp_iter */
    0,                     /* tp_iternext */
    NUECA_methods,             /* tp_methods */
    0,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)NUECA_init,      /* tp_init */
    0,                         /* tp_alloc */
    PyType_GenericNew,         /* tp_new */
};

static PyMethodDef module_methods[] = {
        {NULL}  /* Sentinel */
};

#ifndef PyMODINIT_FUNC  /* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif

PyMODINIT_FUNC initeca(void) {
    PyObject* m;

    if (PyType_Ready(&ECAType) < 0 || PyType_Ready(&NUECAType)) {
        return;
    }

    m = Py_InitModule3("eca", module_methods,
                       "Elementary cellular automata implemented in C.");

    Py_INCREF(&ECAType);
    PyModule_AddObject(m, "ECA", (PyObject *)&ECAType);
    Py_INCREF(&NUECAType);
    PyModule_AddObject(m, "NUECA", (PyObject *)&NUECAType);
}
