from __future__ import division

import log
import eca

from sklearn import linear_model
import numpy as np

import random
import math
import argparse
from collections import deque

class Reservoir(object):
    def __init__(self, R, I, input_size, ca_size, rule1, rule2, inputs):
        self.I = I
        self.R = R
        self.ca_size = ca_size
        self.rule1 = rule1
        self.rule2 = rule2

        self.mappings = []
        self.automatas = []
        for r in xrange(R):
            self.mappings.append(random.sample(xrange(input_size), inputs))
            self.automatas.append(eca.NUECA(self.ca_size, self.rule1, self.rule2))

    def add_values(self, input):
        for r in xrange(self.R):
            for mapping in self.mappings[r]:
                self.automatas[r][mapping] = input

    def get_feature_vector(self):
        output = []

        for i in xrange(self.I):
            for a in self.automatas:
                a.nextState()
                output.extend(a.getState())

        return output

    def reset(self):
        self.automatas = []
        for r in xrange(R):
            self.automatas.append(eca.NUECA(self.ca_size, self.rule1, self.rule2))

    def execute(self, input):
        output = []

        self.add_values(input)
        return self.get_feature_vector()

logger = log.setup_logger()
parser = argparse.ArgumentParser(description = 'Bit parity AND bit density problem')
parser.add_argument('-r', default=8, type=int)
parser.add_argument('-i', default=8, type=int)
parser.add_argument('-n', default=5, type=int, help='How many time steps bit parity should be calculated for')
parser.add_argument('--rule1', default=27, type=int, help='The CA rule for the left part of the CA. Note that the input area always start at cell 0.')
parser.add_argument('--rule2', default=83, type=int)
parser.add_argument('--ca-size', default=40, type=int)
parser.add_argument('--inputs', default=10, type=int, help='How many copies of the input to put on the CA')
parser.add_argument('--input-area', default=20, type=int)
parser.add_argument('--density-delay', default=3, type=int, help='The delay that should be used for the density calculation')

args = parser.parse_args()

R = args.r
I = args.i
N = args.n
ca_size = args.ca_size
rule1 = args.rule1
rule2 = args.rule2
inputs = args.inputs
input_area = args.input_area
density_delay = args.density_delay

# 50 for N=1, which we don't test
test_size = 150
# 50 for N=1, 150 for N=3, 300 for N=5 and 500 for N > 5
train_size = 150
if N == 5:
    train_size = 300
elif N > 5:
    train_size = 500

assert inputs <= input_area <= ca_size

logger.info('Running temporal parity and density with paramteters R = %d, I = %d, N = %d, CA size %d and CA Rule %d and %d' % (R, I, N, ca_size, rule1, rule2))
logger.info('Train data set size: %d, test data set size: %d. Each bit is mapped to %d locations, on a %d grid. Temporal density delayed by %d.' % (train_size, test_size, inputs, input_area, density_delay))

node = Reservoir(R, I, input_area, ca_size, rule1, rule2, inputs)
parityRegNode = linear_model.Ridge()
densityRegNode = linear_model.Ridge()

fit_X = []
fit_parity_y = []
fit_density_y = []
last_bits = deque([0] * N)
history = deque([deque([0] * N) for _ in xrange(density_delay)])

for _ in xrange(train_size):
    input = random.randint(0,1)
    last_bits.append(input)
    last_bits.popleft()

    history.append(deque(last_bits))
    delayed_last_bits = history.popleft()

    fit_X.append(node.execute(input))
    fit_parity_y.append([last_bits.count(1) % 2])
    fit_density_y.append([int(delayed_last_bits.count(1) > delayed_last_bits.count(0))])

parityRegNode.fit(np.array(fit_X), np.array(fit_parity_y))
densityRegNode.fit(np.array(fit_X), np.array(fit_density_y))

node.reset()
last_bits = deque([0] * N)
history = deque([deque([0] * N) for _ in xrange(density_delay)])
parity_correct = 0
density_correct = 0
for _ in xrange(test_size):
    input = random.randint(0,1)
    last_bits.append(input)
    last_bits.popleft()

    history.append(deque(last_bits))
    delayed_last_bits = history.popleft()

    node_output = np.array([node.execute(input)])

    expected = last_bits.count(1) % 2
    out = parityRegNode.predict(node_output)[0]

    if abs(out - expected) < 0.5:
        parity_correct += 1

    expected = int(delayed_last_bits.count(1) > delayed_last_bits.count(0))
    out = densityRegNode.predict(node_output)[0]

    if abs(out - expected) < 0.5:
        density_correct += 1

logger.info("Test completed, parity correctly classified: %f, density correctly classified: %f" % (parity_correct / test_size, density_correct / test_size))
