from __future__ import division

import log
import eca

from sklearn import linear_model
import numpy as np

import random
import math
import argparse
from collections import deque

class Reservoir(object):
    def __init__(self, R, I, input_size, ca_size, ca_rule, inputs):
        self.I = I
        self.R = R
        self.ca_size = ca_size
        self.ca_rule = ca_rule

        self.mappings = []
        self.automatas = []
        for r in xrange(R):
            self.mappings.append(random.sample(xrange(input_size), inputs))
            self.automatas.append(eca.ECA(self.ca_size, self.ca_rule))

    def add_values(self, input):
        for r in xrange(self.R):
            for mapping in self.mappings[r]:
                self.automatas[r][mapping] = input

    def get_feature_vector(self):
        output = []

        for i in xrange(self.I):
            for a in self.automatas:
                a.nextState()
                output.extend(a.getState())

        return output

    def reset(self):
        self.automatas = []
        for r in xrange(R):
            self.automatas.append(eca.ECA(self.ca_size, self.ca_rule))

    def execute(self, input):
        output = []

        self.add_values(input)
        return self.get_feature_vector()

logger = log.setup_logger()
parser = argparse.ArgumentParser(description = 'Bit parity problem')
parser.add_argument('-r', default=8, type=int)
parser.add_argument('-i', default=8, type=int)
parser.add_argument('-n', default=5, type=int, help='How many time steps bit parity should be calculated for')
parser.add_argument('--ca-rule', default=90, type=int)
parser.add_argument('--ca-size', default=40, type=int)
parser.add_argument('--inputs', default=10, type=int, help='How many copies of the input to put on the CA')
parser.add_argument('--input-area', default=20, type=int)

args = parser.parse_args()

R = args.r
I = args.i
N = args.n
ca_size = args.ca_size
ca_rule = args.ca_rule
inputs = args.inputs
input_area = args.input_area

# 50 for N=1, which we don't test
test_size = 150
# 50 for N=1, 150 for N=3, 300 for N=5 and 500 for N > 5
train_size = 150
if N == 5:
    train_size = 300
elif N > 5:
    train_size = 500

assert inputs <= input_area <= ca_size

logger.info('Running temporal parity with paramteters N = %d, R = %d, I = %d, CA size %d and CA Rule %d' % (N, R, I, ca_size, ca_rule))
logger.info('Train data set size: %d, test data set size: %d. Each bit is mapped to %d locations, on a %d grid' % (train_size, test_size, inputs, input_area))

node = Reservoir(R, I, input_area, ca_size, ca_rule, inputs)
regNode = linear_model.Ridge()

fit_X = []
fit_y = []
last_bits = deque([0] * N)

for _ in xrange(train_size):
    input = random.randint(0,1)
    last_bits.append(input)
    last_bits.popleft()

    fit_X.append(node.execute(input))
    fit_y.append([last_bits.count(1) % 2])

regNode.fit(np.array(fit_X), np.array(fit_y))

node.reset()
last_bits = deque([0] * N)
correct = 0
for _ in xrange(test_size):
    input = random.randint(0,1)
    last_bits.append(input)
    last_bits.popleft()

    expected = last_bits.count(1) % 2
    out = regNode.predict(np.array([node.execute(input)]))[0]

    if abs(out - expected) < 0.5:
        correct += 1

logger.info("Test completed, correctly classified: %f" % (correct / test_size))
